Unit Network;

{$X+}
{$MODE Delphi}

Interface

Uses
    Go32, Objects, SysUtils, Util;

Type
	{ 6 Byte ethernet address }
    PEthernetAddress = ^TEthernetAddress;
    TEthernetAddress = packed array[0..5] of Byte;

const
	ETH_BROADCASTADDRESS: TEthernetAddress = ($FF, $FF, $FF, $FF, $FF, $FF);
    ETH_NULLADDRESS: TEthernetAddress = (0, 0, 0, 0, 0, 0);

Type
    { Ethernet Header }
    TEthernetHeader = packed record
		TargetAddr: TEthernetAddress;
		SourceAddr: TEthernetAddress;
		TypeLen   : Word;
    end;

    { Ethernet Packet }
    PEthernetPacket = ^TEthernetPacket;
    TEthernetPacket = packed record
    	Header: TEthernetHeader;
        Data: packed Array[0..1500] of BYTE;
    end;

    TNetworkAdapter = class;
    TNetRoutine = Procedure(Adapter: TNetworkAdapter; Packet: PEthernetPacket);

    TNetIdleRoutine = Procedure;

	TAdapterStatus = (asOk, asError);

	{ Adapter Object }
	TNetworkAdapter = class
	private
		FProtocols: PCollection;
        Procedure   Process;
    protected
        FStatus: TAdapterStatus;
        FAdapterName: AnsiString;
        procedure   CheckStatus; virtual; abstract;
        Function  	IntAddProtocol(TypeLen: WORD): Boolean; virtual; abstract;
        Function    IntRemoveProtocol(TypeLen: WORD): Boolean; virtual; abstract;
        Function	GetPacket(Var Packet: TEthernetPacket): Boolean; virtual; abstract;
    public
        constructor Create;
		destructor  Destroy; override;
        Function    GetAddress(var Addr: TEthernetAddress): Boolean; virtual; abstract;
		Function    SendPkt(Buffer: Pointer; BufLen: Word): Boolean; virtual; abstract;
		Function    AddProtocol(TypeLen: WORD; Routine: TNetRoutine): Boolean; virtual;
        Function    RemoveProtocol(TypeLen: WORD; Routine: TNetRoutine): Boolean; virtual;
        property    Status: TAdapterStatus read FStatus;
        property    AdapterName: AnsiString Read FAdapterName;
	end;

Function  Ethernet_AddrToString(const EtherAddr : TEthernetAddress) : AnsiString;
Function  Ethernet_Compare(const Addr1, Addr2: TEthernetAddress): Boolean;
Function  GetMemCheck(var P: Pointer; Size: WORD): Boolean;
procedure NET_Work_Low;
procedure NET_Work;
procedure NET_Delay(MS: WORD);
procedure NET_AddIdleProc(Proc: TNetIdleRoutine);
procedure NET_RemoveIdleProc(Proc: TNetIdleRoutine);

implementation

Type
	PProtocolEntry = ^TProtocolEntry;
	TProtocolEntry = record
		TypeLen: WORD;
		Routine: TNetRoutine;
	end;

Var
	AdapterList: PCollection;
    IdleProcList: PCollection;




procedure NET_AddIdleProc(Proc: TNetIdleRoutine);
begin
    if IdleProcList.IndexOf(@Proc) <> -1 then IdleProcList^.Delete(@Proc);
    IdleProcList^.Insert(@Proc);
end;

procedure NET_RemoveIdleProc(Proc: TNetIdleRoutine);
begin
    IdleProcList^.Delete(@Proc);
end;

constructor TNetworkAdapter.Create;
begin
    Inherited Create;
    FAdapterName := 'None';
	{ Create Protocol List. It holds the protocols bound to this adapter }
	FProtocols := New(PCollection, Init(10, 1));
	{ Include this adapter in the adapter list }
	AdapterList^.Insert(Self);
    { Check for an installed packet driver }
    CheckStatus;
end;

Destructor TNetworkAdapter.Destroy;
Var
    PE: PProtocolEntry;
begin
    { Quitamos este adaptador de la lista de adaptadores }
    AdapterList^.Delete(Self);
    { Eliminamos los procesadores }
    While FProtocols^.Count > 0 do begin
        PE := FProtocols^.At(0);
        RemoveProtocol(PE.TypeLen, PE.Routine);
    end;
    Dispose(FProtocols, Done);
    { Listo }
    inherited Destroy;
end;

Procedure TNetworkAdapter.Process;
Var
	i: Integer;
	Pkt, PktTmp: TEthernetPacket;
begin
    if (FStatus <> asOk) or not GetPacket(Pkt) then Exit;
    { Hay al menos un paquete. Lo descomponemos }
    for i := 0 to FProtocols^.Count - 1 do begin
        if PProtocolEntry(FProtocols^.At(i))^.TypeLen = Pkt.Header.TypeLen then begin
            PktTmp := Pkt;
            PProtocolEntry(FProtocols^.At(i))^.Routine(Self, @PktTmp);
        end;
    end;
end;

Function TNetworkAdapter.AddProtocol(TypeLen: WORD; Routine: TNetRoutine): Boolean;
Var
    PE: PProtocolEntry;
begin
	AddProtocol := False;
	if Status <> asOk then Exit;
    { First, try to register this protocol with the Packet Driver          }
	if not IntAddProtocol(TypeLen) then Exit;
	{ Done with the Packet Driver. Include the new protocol in the list of }
	{ bound protocols.													   }
	PE := New(PProtocolEntry);
    PE^.TypeLen := TypeLen;
    PE^.Routine := Routine;
    FProtocols^.Insert(PE);
    AddProtocol := True;
end;

Function TNetworkAdapter.RemoveProtocol(TypeLen: WORD; Routine: TNetRoutine): Boolean;
Var
	i: Integer;
    PE: PProtocolEntry;
begin
    RemoveProtocol := False;
    for i := 0 to FProtocols^.Count - 1 do begin
        PE := FProtocols^.At(i);
        if (PE^.TypeLen = TypeLen) and (@PE^.Routine = @Routine) then begin
            IntRemoveProtocol(PE^.TypeLen);
            FProtocols^.Delete(PE);
            Dispose(PE);
            RemoveProtocol := True;
            Break;
        end;
    end;
end;

procedure NET_Delay(MS: WORD);
Var
    Timeout: Comp;
begin
	Timeout := GetMSCounter + MS;
	While GetMSCounter < Timeout do NET_Work_Low;
end;

procedure NET_Work;
Var
	i: Integer;
	IdleProc: TNetIdleRoutine;
begin
	Net_Work_Low;
	for i := 0 to IdleProcList.Count - 1 do begin
		@IdleProc := IdleProcList^.At(i);
		IdleProc;
	end;
end;

procedure NET_Work_Low;
Var
	i: Integer;
begin
    for i := 0 to AdapterList^.Count - 1 do begin
        TNetworkAdapter(AdapterList^.At(i)).Process;
    end;
end;

Function Ethernet_AddrToString(const EtherAddr: TEthernetAddress): AnsiString;
var
	tCount  : Byte;
	tString : AnsiString;
begin
	tString := '';
    for tCount := 0 to 5 do begin
		tString := tString + IntToHex(EtherAddr[tCount], 2) + ':';
    end;
	Ethernet_AddrToString := Copy(tString, 1, 17);
end;

function Ethernet_Compare(const Addr1, Addr2: TEthernetAddress): Boolean;
begin
	Ethernet_Compare :=
      (Addr1[0] = Addr2[0])
      and (Addr1[1] = Addr2[1])
      and (Addr1[2] = Addr2[2])
      and (Addr1[3] = Addr2[3])
      and (Addr1[4] = Addr2[4])
      and (Addr1[5] = Addr2[5]);
end;

Function GetMemCheck(var P: Pointer; Size: WORD): Boolean;
begin
    if MaxAvail >= Size then begin
    	GetMem(P, Size);
       	GetMemCheck := True;
	end else begin
		GetMemCheck := False;
	end;
end;



initialization
	// Create various Objetcs
	AdapterList := New(PCollection, Init(10, 1));
	IdleProcList := New(PCollection, Init(10, 1));
finalization
	While IdleProcList^.Count > 0 do IdleProcList^.AtDelete(0);
	Dispose(IdleProcList, Done);
	Dispose(AdapterList, Done);
end.