Unit WebSrv;

Interface

Uses
    Objects, Crt, SysUtils, Network, Sockets, DosSock, IP, Util;

Type
    TWebMethodType = (wmNone, wmGet, wmPost);

    TWebServer = class;

    TWebRequest = class
    private
    	FSocket: LongInt;
        FURL: AnsiString;
        FQuery: AnsiString;
        FServer: TWebServer;
        FHeader: AnsiString;
        FMethod: AnsiString;
        FContent: AnsiString;
        FPathInfo: AnsiString;
        FQueryFields: AnsiString;
        FMethodType: TWebMethodType;
        Function ReadHeader: Boolean;
        Function GetContentLength: Integer;
        Function GetContentType: AnsiString;
        Function GetQueryFields(Value: AnsiString): AnsiString;
        Function GetHeaderFields(Value: AnsiString): AnsiString;
    public
        constructor Create(Server: TWebServer; HSocket: Integer);
        Function ReadContent: Boolean;
        Function SaveContent(FileName: AnsiString): Boolean;
        property Content: AnsiString Read FContent;
        property ContentType: AnsiString Read GetContentType;
        property ContentLength: Integer Read GetContentLength;
        property QueryFields[Name: AnsiString]: AnsiString Read GetQueryFields;
        property HeaderFields[Name: AnsiString]: AnsiString Read GetHeaderFields;
        property PathInfo: AnsiString Read FPathInfo;
        property Query: AnsiString Read FQuery;
        property URL: AnsiString Read FURL;
        property Method: AnsiString Read FMethod;
        property MethodType: TWebMethodType Read FMethodType;
    end;

    TWebResponse = class
    private
        FSent: Boolean;
        FSocket: LongInt;
        FServer: TWebServer;
        FStatusCode: Integer;
        FContent: AnsiString;
        FHeaderSent: Boolean;
        FLocation: AnsiString;
        FContentLength: Integer;
        FContentType: AnsiString;
    protected
    	procedure SendHeader;
    public
        constructor Create(Server: TWebServer; HSocket: Integer);
        property Sent: Boolean read FSent;
        function SendResponse: Boolean;
        function SendString(s: Ansistring): Boolean;
        function SendFile(FileName: Ansistring): Boolean;
        procedure Redirect(URL: AnsiString);
        property Content: AnsiString Read FContent Write FContent;
        property ContentType: AnsiString Read FContentType Write FContentType;
        property ContentLength: Integer Read FContentLength Write FContentLength;
        property StatusCode: Integer Read FStatusCode Write FStatusCode;
        property Location: AnsiString Read FLocation Write FLocation;
    end;

    TWebProc = procedure(Server: TWebServer; Request: TWebRequest; Response: TWebResponse);

    TActionStruct = record
        Proc: TWebProc;
        PathInfo: string;
    end;

    TWebServer = class
    private
    	FSocket: LongInt;
        FRootPath: AnsiString;
        FDefaultPage: AnsiString;
        FConnections: Array[1..100] of LongInt;
        FActions: Array[1..100] of TActionStruct;
    protected
    	procedure Work; virtual;
        function AddConnection(HSocket: LongInt): Boolean; virtual;
        procedure ProcessAction(Request: TWebRequest; Response: TWebResponse); virtual;
        procedure ProcessConnection(ConnIndex: Integer); virtual;
    public
        Constructor Create;
        Destructor Destroy; override;
        function Start(Address: TIPAddress; Port: WORD; RootPath, DefaultPage: AnsiString): Boolean;
        function AddAction(Action: AnsiString; Proc: TWebProc): Boolean;
        property RootPath: AnsiString Read FRootPath;
        property DefaultPage: AnsiString Read FDefaultPage;
    end;


implementation

Type
    TContType = record
        Ext: AnsiString;
        ContentType: AnsiString;
    end;

Const
    ContTypes: Array[1..10] of TContType = (
      (Ext: '.TXT'; ContentType: 'text/plain'),
      (Ext: '.DOC'; ContentType: 'application/msword'),
      (Ext: '.BAT'; ContentType: 'text/plain'),
      (Ext: '.INI'; ContentType: 'text/plain'),
      (Ext: '.HTM'; ContentType: 'text/html'),
      (Ext: '.EXE'; ContentType: 'application/octet-stream'),
      (Ext: '.COM'; ContentType: 'application/octet-stream'),
      (Ext: '.GIF'; ContentType: 'image/gif'),
      (Ext: '.JPG'; ContentType: 'image/jpeg'),
      (Ext: '.BMP'; ContentType: 'image/bmp')
      );

Var
    ServerList: PCollection;

function Web_GetParam(const Params, ParamName, Default: AnsiString): AnsiString;
Var
    i: Integer;
    S: AnsiString;
begin
	Result := Default;
    for i := 1 to 1000 do begin
        S := ParseParamByNumber(Params, i, '&');
        if S = '' then Exit;
        if UpperCase(ParseParamByNumber(S, 1, '=')) = UpperCase(ParamName) then begin
            Result := ParseParamByNumber(S, 2, '=');
            Exit;
        end;
    end;
end;

function Web_GetContentType(FileName: AnsiString): AnsiString;
Var
    i: Integer;
    Ext: AnsiString;
begin
    Ext := UpperCase(ExtractFileExt(FileName));
    for i := Low(ContTypes) to High(ContTypes) do begin
        if ContTypes[i].Ext = Ext then begin
            Result := ContTypes[i].ContentType;
            Exit;
        end;
    end;
    Result := 'application/octet-stream';
end;

{ TBebRequest }

constructor TWebRequest.Create(Server: TWebServer; HSocket: Integer);
begin
    inherited Create;
    FHeader  := '';
    FContent := '';
    FServer  := Server;
    FSocket  := HSocket;
    ReadHeader;
end;

Function TWebRequest.GetQueryFields(Value: AnsiString): AnsiString;
begin
    Result := Web_GetParam(FQuery, Value, '');
end;

Function TWebRequest.GetHeaderFields(Value: AnsiString): AnsiString;
Var
    i: Integer;
    HeaderLine: AnsiString;
begin
	Value := UpperCase(Value);
    for i := 1 to 1000 do begin
        HeaderLine := ParseParamByNumber(FHeader, i, CRLF);
        if HeaderLine = '' then break;
        if UpperCase(ParseParamByNumber(HeaderLine, 1, ':')) = Value then begin
            Result := ParseParamByNumber(HeaderLine, 2, ':');
            Exit;
        end;
    end;
    Result := '';
end;

Function TWebRequest.GetContentType: AnsiString;
begin
    Result := GetHeaderFields('Content-Type');
end;

Function TWebRequest.GetContentLength: AnsiString;
begin
    Result := IVal(GetHeaderFields('Content-Length'));
end;

Function TWebRequest.SaveContent(FileName: AnsiString): Boolean;
Var
    F: File;
    L: LongInt;
    Timeout: Comp;
    Cnt, i: Integer;
    Buffer: Array[1..1024] of char;
begin
    // Create the output file
    Assign(F, FileName);
	try
        ReWrite(F, 1);
    except
        Result := False;
        Exit;
    end;
    // Receive Content, with a timeout of 10 seconds between chunks
    Cnt := 0;
    Timeout := GetMSCounter + 10000;
    While Cnt < ContentLength do begin
        if recv(FSocket, buffer, 0, 0) <> 0 then break;
        IOCtlSocket(FSocket, FIONREAD, @L);
        if L > 0 then begin
//        	WriteLn(L);
            if L > SizeOf(buffer) then L := SizeOf(Buffer);
            L := recv(FSocket, buffer, L, 0);
            try
                BlockWrite(f, buffer, L);
            except
            	Break;
            end;
            Inc(Cnt, L);
            Timeout := GetMSCounter + 10000;
        end;
    end;
    Close(f);
    if (Cnt = ContentLength) then
        Result := True
    else begin
        Erase(f);
        Result := False;
    end;
end;

Function TWebRequest.ReadContent: Boolean;
begin
    if SaveContent('__CNT.TMP') then begin
    	FContent := FileToString('__CNT.TMP');
        DeleteFile('__CNT.TMP');
        Result := True;
    end else begin
    	Result := False;
    end;
end;

function TWebRequest.ReadHeader: Boolean;
Var
    c: char;
    L: LongInt;
    i: Integer;
    Timeout: Comp;
    Line, Params, FilePath: AnsiString;
begin
    // Get Header by reading from the socket.
    Result := False;
    FHeader := '';
    Timeout := GetMSCounter + 5000;
    While not Result and (GetMSCounter < Timeout)
      and (recv(FSocket, c, 0, 0) = 0) do begin
        ioctlSocket(FSocket, FIONREAD, @L);
        if L > 0 then begin
            recv(FSocket, c, 1, 0);
            FHeader := FHeader + c;
        end;
        Result := Pos(CRLF + CRLF, FHeader) <> 0;
    end;
    if not Result then Exit;

//    WriteLn('Content-Type  : ', ContentType);
//    WriteLn('Content-Length: ', IntToStr(ContentLength));

    // Extract basic information
    Line    := ParseParamByNumber(FHeader, 1, CRLF);
    FMethod := ParseParamByNumber(Line, 1, ' ');
    FURL    := ParseParamByNumber(Line, 2, ' ');

    // Check method type
    if UpperCase(FMethod) = 'GET' then
        FMethodType := wmGet
    else if UpperCase(FMethod) = 'POST' then
        FMethodType := wmPost
    else begin
        FMethodType := wmNone;
        Exit;
    end;

    // Check URL, PathInfo and Query
    i := Pos('?', FURL);
    if i = 0 then begin
        FQuery := '';
        FPathInfo := FURL;
    end else begin
        FQuery := Copy(FURL, i + 1, Length(FURL));
        FPathInfo := Copy(FURL, 1, i - 1);
    end;
//	WriteLn(Method, ',', URL, ',', PathInfo, ',', Query);

    // Check for default Page
    if FPathInfo = '/' then FPathInfo := FServer.DefaultPage;
end;

{ TWebResponse }
constructor TWebResponse.Create(Server: TWebServer; HSocket: Integer);
begin
    inherited Create;
    FSent := False;
    FContent := '';
    FServer := Server;
    FSocket := HSocket;
    FLocation := '';
    FStatusCode := 200;
    FHeaderSent := False;
    FContentLength := -1;
    FContentType := 'text/html';
end;

procedure TWebResponse.SendHeader;
Var
	Str: AnsiString;
begin
    Str :=
      'HTTP/1.1 ' + IntToStr(StatusCode) + ' ' + CRLF +
      'Server: FreePascal Free Server/1.1 by Diego De Marco' + CRLF +
      'Content-Type: ' + ContentType + CRLF;
    if FContentLength >= 0 then Str := Str + 'Content-Length: ' +
      IntToStr(FContentLength) + CRLF;
    if FLocation <> '' then Str := Str + 'Location: ' + FLocation + CRLF;
    Str := Str + CRLF;
    send(FSocket, Str[1], Length(Str), 0);
    FHeaderSent := True;
end;

function TWebResponse.SendResponse: Boolean;
begin
    if FContent = '' then begin
        if not FHeaderSent then SendHeader;
        Result := True;
    end else begin
        if not FHeaderSent then FContentLength := Length(FContent);
        Result := SendString(Content);
    end;
end;

function TWebResponse.SendString(s: Ansistring): Boolean;
begin
    if Length(s) = 0 then
        Result := True
    else begin
        if not FHeaderSent then SendHeader;
        Result := send(FSocket, s[1], Length(s), 0) = Length(s);
    end;
end;

function TWebResponse.SendFile(FileName: Ansistring): Boolean;
Var
	Buffer: AnsiString;
begin
    Buffer := FileToString(FileName);
    ContentType := Web_GetContentType(FileName);
    ContentLength := Length(Buffer);
    Result := SendString(Buffer);
end;

procedure TWebResponse.Redirect(URL: AnsiString);
begin
    StatusCode := 302;
    Location := URL;
end;

{ TWebServer }

Constructor TWebServer.Create;
Var
	i: Integer;
begin
    inherited Create;
    FRootPath := '';
    FDefaultPage := '';
    FSocket := 0;
    for i := Low(FConnections) to High(FConnections) do FConnections[i] := 0;
    for i := Low(FActions) to High(FActions) do begin
        FActions[i].PathInfo := '';
        @FActions[i].Proc := nil;
   	end;
    ServerList.Insert(Self);
end;

Destructor TWebServer.Destroy;
Var
	i: Integer;
begin
    ServerList.Delete(Self);
    if FSocket <> 0 then CloseSocket(FSocket);
    for i := Low(FConnections) to High(FConnections) do begin
        if FConnections[i] <> 0 then begin
            CloseSocket(FConnections[i]);
            FConnections[i] := 0;
        end;
    end;
	inherited Destroy;
end;

function TWebServer.Start(Address: TIPAddress; Port: WORD;
  RootPath, DefaultPage: AnsiString): Boolean;
Var
    Addr: TInetSockAddr;
begin
    Result := False;
    // Open Server Socket to accept incoming connections
	Addr.Family   := PF_INET;
    Addr.Port     := Port;
    Addr.Addr     := Address;
    FSocket 	  := Socket(PF_INET, SOCK_STREAM, IPPROTO_IP);
    if (FSocket <= 0) or not Bind(FSocket, Addr, SizeOf(Addr))
      or not Listen(FSocket, 10) then begin
        if FSocket <> 0 then CloseSocket(FSocket);
        FSocket := 0;
        Exit;
	end;
    // Done
    FRootPath := RootPath;
    FDefaultPage := DefaultPage;
    Result := True;
end;

function TWebServer.AddConnection(HSocket: LongInt): Boolean;
Var
	i: Integer;
begin
    for i := Low(FConnections) to High(FConnections) do begin
        if FConnections[i] = 0 then begin
            FConnections[i] := HSocket;
            Result := True;
            Exit;
        end;
    end;
    Result := False;
end;

function TWebServer.AddAction(Action: AnsiString; Proc: TWebProc): Boolean;
Var
	i: Integer;
begin
    for i := Low(FActions) to High(FActions) do begin
        if @FActions[i].Proc = nil then begin
            FActions[i].PathInfo := Action;
            FActions[i].Proc := Proc;
        	Result := True;
        	Exit;
        end;
    end;
    Result := False;
end;

procedure TWebServer.ProcessAction(Request: TWebRequest; Response: TWebResponse);
Var
    i: Integer;
    Path: AnsiString;
begin
	// Check for a handler registered for this request.
    for i := Low(FActions) to High(FActions) do begin
        if (Uppercase(FActions[i].PathInfo) = Uppercase(Request.PathInfo))
          and Assigned(@FActions[i].Proc) then begin
            FActions[i].Proc(Self, Request, Response);
            Exit;
        end;
    end;

    // There was no handler. Check to see if it's a file.
    Path := Copy(Request.PathInfo, 2, Length(Request.PathInfo));
    for i := 1 to Length(Path) do if Path[i] = '/' then Path[i] := '\';
    Path := GoodDir(RootPath) + Path;
    if FileExists(Path) then begin
        Response.SendFile(Path);
        Exit;
    end;

    // There was no handler or file. Show Error 404 "Not Found".
    Response.StatusCode := 404;
    Response.Content :=
      '<HTML><TITLE>404 NOT FOUND</TITLE><H1><B>404 NOT FOUND</B></H1>' +
      'The file ' + Request.PathInfo + ' is not on this server.<br>' +
      'Please try again using another URL.' +
      '</HTML>';
end;

procedure TWebServer.ProcessConnection(ConnIndex: Integer);
Var
	c: char;
    ToRead: LongInt;
    HSocket: LongInt;
    Request: TWebRequest;
    Response: TWebResponse;
begin
    HSocket := FConnections[ConnIndex];
    // Test for incoming request
    IOCtlSocket(HSocket, FIONREAD, @ToRead);
    if ToRead = 0 then Exit;
    // Process incoming request
//  Write('Processing... ');
    Request := TWebRequest.Create(Self, HSocket);
    Response := TWebResponse.Create(Self, HSocket);
//	WriteLn(Request.FHeader);
    try
        if Request.MethodType <> wmNone then begin
            ProcessAction(Request, Response);
            if not Response.Sent then Response.SendResponse;
        end;
    finally
        Request.Free;
        Response.Free;
        CloseSocket(HSocket);
 		FConnections[ConnIndex] := 0;
    end;
//  WriteLn('Done.');
end;

procedure TWebServer.Work;
Var
    HSock, L: LongInt;
    ConnIndex: Integer;
    Addr: TInetSockAddr;
begin
    if FSocket = 0 then Exit;
    // Accept next incomming connection
    L := SizeOf(Addr);
    HSock := Accept(FSocket, Addr, L);
    if HSock > 0 then AddConnection(HSock);
    // Process active connections
    for ConnIndex := Low(FConnections) to High(FConnections) do begin
        if FConnections[ConnIndex] <> 0 then begin
            ProcessConnection(ConnIndex);
            NET_Work_Low;
        end;
    end;
end;


procedure Web_Work;
Var
	i: Integer;
begin
    NET_Work_Low;
    for i := 0 to ServerList.Count - 1 do begin
        TWebServer(ServerList.At(i)).Work;
    end;
end;


Var
    Server, i: Integer;
initialization
    // Creamos la lista de Servers
    New(ServerList, Init(10, 1));
    // Agregamos un procedure de Idle
	NET_AddIdleProc(@Web_Work);
finalization
	// Quitamos el procedure de Idle
    NET_RemoveIdleProc(@Web_Work);
    // Quitamos los Servers
    for i := 0 to ServerList.Count - 1 do TWebServer(ServerList.At(i)).Free;
    // Eliminamos la lista de servers
    Dispose(ServerList, Done);
end.